from sre_compile import dis
from typing import List
import os

import re
import json
from fuzzywuzzy import fuzz

from kalapaocr.config import KEYWORDS, DATA_CRAWL


def check_key_belong_to_line(key, words, thres=0.8):
    len_key = len(key.split(" "))
    len_words = len(words)
    if len_key > len_words:
        return words, False

    for i in range(len_words - len_key + 1):
        consider_words = words[i : i + len_key]
        consider_key = " ".join([x.strip() for x in consider_words]).strip()
        similarity_ratio = fuzz.ratio(key.lower(), consider_key.lower()) / 100.0
        if similarity_ratio > thres and similarity_ratio < 1:
            words_of_key = key.split(" ")
            for j in range(i, i + len_key):
                words[j] = words_of_key[j - i]
            return words, True
    return words, False


def postprocess_ocr_with_keywords(
    text_line: str, keywords: List[str] = KEYWORDS, thres: float = 0.85
):
    words = re.split(r"\s|,|-|:|;", text_line)
    for key in keywords:
        words, flag = check_key_belong_to_line(key=key, words=words, thres=thres)
    return " ".join([w.strip() for w in words]).strip()


def get_info_name_type_unsignedname(data: List):
    res = {}
    res["name"] = preprocess_text(data[1])
    res["type"] = data[2]
    res["unsign_name"] = data[3]
    return res


def get_street_for_district(street_parser, province_name: str, district_name: str):
    idx_dict = street_parser
    province_name = revert_preprocess_text(province_name)
    district_name = revert_preprocess_text(district_name)
    if idx_dict.get(province_name) == None:
        return []
    province_dict = idx_dict[province_name]
    district_list = province_dict["district"]
    for district in district_list:
        if district["name"] == district_name:
            return district["street"]
    return []


def parser_geo_json(data: dict = DATA_CRAWL, street_parser_path: str = None):
    street_parser = json.load(open(street_parser_path))
    parser_data = {}
    parser_data["provinces"] = []
    for province_info in data:
        province_dict = get_info_name_type_unsignedname(data=province_info)
        districts = province_info[4]
        province_dict["districts"] = []
        district_names = []
        for district in districts:
            district_dict = get_info_name_type_unsignedname(data=district)
            towns = district[4]
            district_dict["towns"] = []
            for town in towns:
                town_dict = get_info_name_type_unsignedname(town)
                district_dict["towns"].append(town_dict)
            district_dict["street"] = get_street_for_district(
                street_parser=street_parser,
                province_name=province_dict["name"],
                district_name=district_dict["name"],
            )
            if district_dict["name"] not in district_names:
                province_dict["districts"].append(district_dict)
                district_names.append(district_dict["name"])
            else:
                for distr in province_dict["districts"]:
                    if distr["name"] == district_dict["name"]:
                        distr["towns"].extend(district_dict["towns"])

        parser_data["provinces"].append(province_dict)
    return parser_data


def revert_preprocess_text(text):
    text = text.replace("uỷ", "ủy")
    text = text.replace("uỵ", "ụy")
    text = text.replace("uỳ", "ùy")
    text = text.replace("uý", "úy")
    text = text.replace("oá", "óa")
    text = text.replace("oà", "òa")
    return text


def preprocess_text(text):
    text = text.replace("ủy", "uỷ")
    text = text.replace("ụy", "uỵ")
    text = text.replace("ùy", "uỳ")
    text = text.replace("úy", "uý")
    text = text.replace("óa", "oá")
    text = text.replace("òa", "oà")
    return text


def check_has_digit_word(text):
    for idx, word in enumerate(text.split(" ")):
        for char in word:
            if char.isdigit():
                return idx
    return -1


def hard_postprocess(text):
    words = text.split(" ")
    words[0] = words[0].replace("Thơn", "Thôn")
    words[0] = words[0].replace("Thồn", "Thôn")
    text = " ".join(words).strip()
    text = text.replace("ủy", "uỷ")
    text = text.replace("ụy", "uỵ")
    text = text.replace("ùy", "uỳ")
    text = text.replace("úy", "uý")
    text = text.replace("óa", "oá")
    text = text.replace("òa", "oà")
    # text = text.replace("Sơn Đông Linh", "Sơn Đông")
    # text = text.replace("Thị Tránh An", "Thị Trấn Vĩnh An")
    return text


def find_pattern(words, patterns):
    max_similar = 0
    match_pattern = None
    for pattern in patterns:
        pattern_name = pattern["name"]
        len_words_of_hint = len(pattern_name.split(" "))
        if len_words_of_hint > len(words):
            continue
        consider_pattern = " ".join(words[-1 * len_words_of_hint :]).strip()
        similar_score = (
            fuzz.ratio(pattern_name.lower(), consider_pattern.lower()) / 100.0
        )

        if consider_pattern.lower() == pattern_name.lower():
            return 1.0, pattern
        if max_similar < similar_score:
            max_similar = similar_score
            match_pattern = pattern
    if match_pattern is not None:
        idx_digit = check_has_digit_word(match_pattern["name"])
        num_word = len(match_pattern["name"].split(" "))
        if idx_digit >= 0:
            is_replaced_words = words[-1 * num_word :]
            if (
                match_pattern["name"].split(" ")[idx_digit]
                != is_replaced_words[idx_digit]
            ):
                return 0, None
    return max_similar, match_pattern


def postprocess_privince(text: str, parser_data: dict):
    new_text = text
    provinces = parser_data["provinces"]
    word_texts = text.split(" ")
    max_similar, match_province = find_pattern(word_texts, provinces)
    if max_similar > 0.85:
        word_texts[-1 * len(match_province["name"].split(" ")) :] = match_province[
            "name"
        ].split(" ")[:]
        new_text = " ".join(word_texts).strip()
        new_text = postprocess_district(
            new_text,
            match_province,
            pop_element_number=len(match_province["name"].split(" ")),
        )
    return new_text


def postprocess_district(text: str, province: dict, pop_element_number: int):
    new_text = text
    words = text.split(" ")
    province_name_split = words[-1 * pop_element_number :]
    words = words[: len(words) - pop_element_number]
    if not words:
        return new_text
    if words[-1].lower() in ["tp", "t"]:
        province_name_split = [words[-1]] + province_name_split
        words = words[:-1]
    elif fuzz.ratio(" ".join(words[-2:]).strip().lower(), "thành phố") > 90:
        province_name_split = words[-2:] + province_name_split
        words = words[:-2]
    elif fuzz.ratio(words[-1].lower(), "tỉnh") > 90:
        province_name_split = [words[-1]] + province_name_split
        words = words[:-1]
    districts = province["districts"]
    max_similar, match_district = find_pattern(words, districts)
    if max_similar >= 0.8:
        words[-1 * len(match_district["name"].split(" ")) :] = match_district[
            "name"
        ].split(" ")[:]
        words = words + province_name_split
        new_text = " ".join(words).strip()
        new_text = postprocess_town(
            new_text,
            match_district,
            pop_element_number=len(match_district["name"].split(" "))
            + len(province_name_split),
        )
    return new_text


def postprocess_town(text: str, district: dict, pop_element_number: int):
    new_text = text
    words = text.split(" ")
    district_province_name_split = words[-1 * pop_element_number :]
    words = words[: len(words) - pop_element_number]
    if not words:
        return new_text
    if fuzz.ratio(" ".join(words[-2:]).strip().lower(), "thành phố") > 90:
        district_province_name_split = words[-2:] + district_province_name_split
        words = words[:-2]
    elif fuzz.ratio(" ".join(words[-2:]).strip().lower(), "thị xã") > 90:
        district_province_name_split = words[-2:] + district_province_name_split
        words = words[:-2]
    elif (
        fuzz.ratio(words[-1].lower(), "huyện") > 90
        or fuzz.ratio(words[-1].lower(), "quận") > 90
    ):
        district_province_name_split = [words[-1]] + district_province_name_split
        words = words[:-1]
    elif words[-1].lower() in ["tp", "q", "h"]:
        district_province_name_split = [words[-1]] + district_province_name_split
        words = words[:-1]
    towns = district["towns"]
    max_similar, match_town = find_pattern(words, towns)
    if (max_similar > 0.8) or (
        max_similar > 0.7 and len(match_town["name"].split(" ")) >= 2
    ):
        words[-1 * len(match_town["name"].split(" ")) :] = match_town["name"].split(
            " "
        )[:]
        words = words + district_province_name_split
        consider_text = " ".join(
            words[
                : -1
                * (
                    len(match_town["name"].split(" "))
                    + len(district_province_name_split)
                )
            ]
        ).strip()
        street_names = [
            preprocess_text(x)
            for x in district["street"]
            if len(x.split(" ")) >= 2 and check_has_digit_word(x) < 0
        ]
        street_name = postprocess_ocr_with_keywords(
            consider_text, keywords=street_names
        )
        new_text = (
            street_name
            + " "
            + match_town["name"]
            + " "
            + " ".join(district_province_name_split).strip()
        )
    else:
        street_names = [
            preprocess_text(x)
            for x in district["street"]
            if len(x.split(" ")) >= 3 and check_has_digit_word(x) < 0
        ]
        consider_text = " ".join(words).strip()
        street_name = postprocess_ocr_with_keywords(
            consider_text, keywords=street_names, thres=0.8
        )

        new_text = street_name + " " + " ".join(district_province_name_split).strip()
    return new_text.strip()



class Vocab:
    def __init__(self, chars):
        self.pad = 0
        self.go = 1
        self.eos = 2
        self.mask_token = 3

        self.chars = chars

        self.c2i = {c: i + 4 for i, c in enumerate(chars)}

        self.i2c = {i + 4: c for i, c in enumerate(chars)}

        self.i2c[0] = "<pad>"
        self.i2c[1] = "<sos>"
        self.i2c[2] = "<eos>"
        self.i2c[3] = "*"

    def encode(self, chars):
        labels = [self.go]
        for char in chars:
            if char in self.c2i:
                labels.append(self.c2i[char])
            else:
                print("{} out of vocab".format(char))
                labels.append(self.pad)
        labels += [self.eos]
        # return [self.go] + [self.c2i[c] for c in chars] + [self.eos]
        return labels

    def decode(self, ids):
        first = 1 if self.go in ids else 0
        last = ids.index(self.eos) if self.eos in ids else None
        sent = ""
        for i in ids[first:last]:
            if i != self.pad:
                sent += self.i2c[i]
            else:
                sent += " "
        # sent = ''.join([self.i2c[i] for i in ids[first:last]])
        return sent

    def __len__(self):
        return len(self.c2i) + 4

    def batch_decode(self, arr):
        texts = [self.decode(ids) for ids in arr]
        return texts

    def __str__(self):
        return self.chars



import json
from typing import *

import numpy as np

from kalapaocr.recognize import OcrEngine
from kalapaocr.tool.postprocess import (
    hard_postprocess,
    parser_geo_json,
    postprocess_ocr_with_keywords,
    postprocess_privince,
)


class TextRecognitor:
    def __init__(
        self,
        cnn_path,
        encoder_path,
        decoder_path,
        vocab=None,
        street_json_path=None,
        debug=False,
        log_dir=None,
    ):
        if vocab is None:
            vocab = "aAàÀảẢãÃáÁạẠăĂằẰẳẲẵẴắẮặẶâÂầẦẩẨẫẪấẤậẬbBcCdDđĐeEèÈẻẺẽẼéÉẹẸêÊềỀểỂễỄếẾệỆfFgGhHiIìÌỉỈĩĨíÍịỊjJkKlLmMnNoOòÒỏỎõÕóÓọỌôÔồỒổỔỗỖốỐộỘơƠờỜởỞỡỠớỚợỢpPqQrRsStTuUùÙủỦũŨúÚụỤưƯừỪửỬữỮứỨựỰvVwWxXyYỳỲỷỶỹỸýÝỵỴzZ0123456789!\"#$%&'()*+,-./:;<=>?@[\]^_`{|}~ "
        self.text_recogition = OcrEngine(
            cnn_path,
            encoder_path,
            decoder_path,
            vocab=vocab,
            debug=debug,
            log_dir=log_dir,
        )
        self.data_parser = parser_geo_json(street_parser_path=street_json_path)

    def __call__(self, image: np.ndarray) -> str:
        """Input list cropped line image and rectangle coordinates perspective
        Output List Dictionary object contains ocr result with live level and word level

        :param image: line image
        :type image: np.ndarray
        :return: ocr result
        :rtype: str
        """
        res = self.text_recogition(
            img=image, image_height=64, image_max_width=1200, image_min_width=32
        )
        res = postprocess_ocr_with_keywords(res)
        res = hard_postprocess(res)
        res = postprocess_privince(res, self.data_parser)
        res = hard_postprocess(res)
        return res



import os

import cv2
import numpy as np
import onnxruntime as rt
import pandas as pd
from PIL import Image

from kalapaocr.tool.img_proc import is_empty_image


def load_graph_onnx(model_path):
    sess_opt = rt.SessionOptions()
    sess_opt.graph_optimization_level = rt.GraphOptimizationLevel.ORT_ENABLE_BASIC
    session = rt.InferenceSession(
        model_path, sess_options=sess_opt, providers=["CUDAExecutionProvider"]
    )
    return session


def bmm(x, y):
    # Reshape the input tensors to have shape (batch_size, num_rows, num_cols)
    x = np.reshape(x, (x.shape[0], x.shape[1], -1))
    y = np.reshape(y, (y.shape[0], -1, y.shape[2]))

    # Compute the batch matrix-matrix product using numpy.matmul
    out = np.matmul(x, y)

    # Reshape the output tensor back to its original shape
    out = np.reshape(out, (out.shape[0], out.shape[1], -1))

    return out


def calculate_attention(image, attention_weights, index, log_dir=None, vis=False):
    batch_size = attention_weights.shape[0]
    height, width = image.shape[-2:]
    T = attention_weights.shape[1] // 4
    ct = [x for x in range(T)]
    for i in range(T):
        if i == 0:
            ct[i] = 2.5
            continue
        if i == T - 1:
            ct[i] = 2 + (i - 1 - 2) * 4 + 6 + ((width - (2 + (i - 1 - 2) * 4 + 6)) / 2)
        else:
            ct[i] = 2 + (i - 1) * 4 + 3
    ct = np.expand_dims(ct, axis=0)
    ct = np.vstack((ct, ct, ct, ct))
    ct = ct.transpose((1, 0))
    ct = ct.flatten()
    batch_ct = np.zeros_like(attention_weights)
    for i in range(batch_size):
        batch_ct[i] = ct
    attention_weights = np.expand_dims(attention_weights, axis=1)
    batch_ct = np.expand_dims(batch_ct, axis=-1)
    centers = bmm(attention_weights, batch_ct).squeeze(1)
    if vis:
        os.makedirs(log_dir, exist_ok=True)
        image = image.squeeze(0) * 255
        image = image.transpose((1, 2, 0))
        center = centers[0]
        cv2.line(
            image,
            (int(center), 0),
            (int(center), height),
            (0, 255, 0),
            thickness=3,
            lineType=8,
        )
        image_path = os.path.join(log_dir, "{}.jpg".format(index))
        cv2.imwrite(image_path, image)
    return centers


def create_submission(paths, results, output_path):
    res = {"id": [], "answer": []}
    for path, s in zip(paths, results):
        relative_path = path.split("/")[-2] + "/" + path.split("/")[-1]
        res["id"].append(relative_path)
        res["answer"].append(s)
    df = pd.DataFrame(res)
    df.to_csv(output_path, index=False)


def resize(w, h, expected_height, image_min_width, image_max_width):
    ratio = expected_height / float(h)
    new_w = int(expected_height * float(w) / float(h))
    # round_to = 10
    # new_w = math.ceil(new_w/round_to)*round_to
    new_w = max(new_w, image_min_width)
    new_w = min(new_w, image_max_width)

    return new_w, expected_height, ratio


def process_image(image, image_height, image_min_width, image_max_width):
    image = Image.fromarray(image).convert("RGB")
    img = image.convert("RGB")

    w, h = img.size
    new_w, image_height, ratio = resize(
        w, h, image_height, image_min_width, image_max_width
    )

    img = img.resize((new_w, image_height))
    img = np.asarray(img).transpose(2, 0, 1)
    img = img / 255
    img = img[np.newaxis, ...]
    return img, ratio


def process_text_level2(text, last_idx_char):
    words = text.split(" ")
    idx_start_word = 0
    for i, word in enumerate(words):
        if i < len(words) - 1:
            word = word + " "
        idx_end_word = idx_start_word + len(word) - 1
        if last_idx_char in range(idx_start_word, idx_end_word + 1):
            return text[: idx_end_word + 1]
        idx_start_word = idx_end_word + 1
    return text[:last_idx_char]


def remove_noise_text(img, text, probs, locations):
    if len(text) <= 3:
        return text
    last_idx_char = -1
    img = img.astype(np.uint8)
    for i in range(2, len(text)):
        coef = 5 if i < 50 else 3.3
        if abs(locations[i] - locations[i - 1]) > coef * abs(
            abs(locations[i - 1] - locations[i - 2])
        ):
            w_var = 64 * 2  # if i < 50 else img.shape[1] - int(locations[i - 1])
            check_var = img[
                :,
                int(locations[i - 1]) : min(
                    int(locations[i - 1]) + w_var, img.shape[1]
                ),
            ]  # 64: h_img
            is_empty = is_empty_image(check_var)
            if is_empty:
                last_idx_char = i - 1
                break

    if last_idx_char >= 0:
        text = process_text_level2(text, last_idx_char)
        # text = text[:last_idx_char]

    return text



import gdown
import os
import argparse


def parse_args():
    parser = argparse.ArgumentParser(description="Down load model from google drive")
    parser.add_argument(
        "--save_dir",
        "-p",
        required=True,
        type=str,
        help="model paths",
    )
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    url_links = {
        "encoder.onnx": "https://drive.usercontent.google.com/download?id=1hPJphUajDdbz7W47D2RvuDe2qImq7xgY&export=download&authuser=0&confirm=t&uuid=7f1e8b1f-b7c3-40f3-a384-acb83d6aad0f&at=APZUnTWCI39CPd98bWjrUqL9o_6Y:1698682706043",
        "decoder.onnx": "https://drive.usercontent.google.com/download?id=1BAUTUoebUBm0f0saILWGaV37pM-lwu-x&export=download&authuser=0&confirm=t&uuid=50372483-def7-4cce-a596-079d763eed54&at=APZUnTUUyRIKK6itjX-LLRBBo1Ul:1698682739555",
        "cnn.onnx": "https://drive.usercontent.google.com/download?id=1y7x9zofI1hFrjg3V3sktivIFuMP3uxoN&export=download&authuser=0&confirm=t&uuid=fda7fd53-6e2d-4948-8d3f-cfd75c2837e1&at=APZUnTXubx2NmK1D6Oq9fE-BTA5B:1698682775739",
    }

    model_dir = args.save_dir
    os.makedirs(model_dir, exist_ok=True)
    for model in url_links:
        path_model = os.path.join(model_dir, model)
        gdown.download(url_links[model], path_model, quiet=False)

    print("Model dirs: ", os.path.abspath(model_dir))


if __name__ == "__main__":
    main()



import yaml


class Cfg(dict):
    def __init__(self, config_dict):
        super(Cfg, self).__init__(**config_dict)
        self.__dict__ = self

    @staticmethod
    def load_config_from_file(fname):
        with open(fname, encoding="utf-8") as f:
            config = yaml.safe_load(f)
        return Cfg(config)

    def save(self, fname):
        with open(fname, "w") as outfile:
            yaml.dump(dict(self), outfile, default_flow_style=False, allow_unicode=True)



import numpy as np

from kalapaocr.tool.img_proc import is_empty_image
from kalapaocr.tool.vocab import Vocab
from kalapaocr.utils import (
    calculate_attention,
    load_graph_onnx,
    process_image,
    remove_noise_text,
)


def softmax(x, axis=-1):
    """Compute softmax values for each sets of scores in x."""
    sum_exp = np.sum(np.exp(x), axis=axis)
    eta = np.ones_like(sum_exp) * (1e-7)
    return np.exp(x) / (sum_exp + eta)


class OcrEngine:
    def __init__(
        self, cnn_path, encoder_path, decoder_path, vocab, debug=False, log_dir=None
    ):
        self.cnn_session = load_graph_onnx(cnn_path)
        self.encoder_session = load_graph_onnx(encoder_path)
        self.decoder_session = load_graph_onnx(decoder_path)
        self.vocab = Vocab(vocab)
        self.session = (self.cnn_session, self.encoder_session, self.decoder_session)
        self.debug = debug
        self.log_dir = log_dir

    def translate_onnx(
        self,
        img,
        session,
        max_seq_length=100,
        sos_token=1,
        eos_token=2,
    ):
        """data: BxCxHxW"""
        cnn_session, encoder_session, decoder_session = session

        # create cnn input
        cnn_input = {cnn_session.get_inputs()[0].name: np.asarray(img, dtype="float16")}
        src = cnn_session.run(None, cnn_input)

        # create encoder input
        encoder_input = {
            encoder_session.get_inputs()[0].name: np.asarray(src[0], dtype="float32")
        }
        encoder_outputs, hidden = encoder_session.run(None, encoder_input)
        translated_sentence = [[sos_token] * len(img)]
        char_probs = [[1] * len(img)]

        max_length = 0
        batch_char_locations = []
        while max_length <= max_seq_length and not all(
            np.any(np.asarray(translated_sentence).T == eos_token, axis=1)
        ):
            tgt_inp = np.array(translated_sentence)
            decoder_input = {
                decoder_session.get_inputs()[0].name: tgt_inp[-1],
                decoder_session.get_inputs()[1].name: hidden,
                decoder_session.get_inputs()[2].name: encoder_outputs,
            }

            output, hidden, attention_weights = decoder_session.run(None, decoder_input)

            output = np.expand_dims(output, axis=1)
            output = softmax(output)
            indices = np.argmax(output, axis=-1)[:, -1]
            indices = indices.tolist()
            values = np.max(output, axis=-1)[:, -1]
            values = values.tolist()
            char_probs.append(values)

            translated_sentence.append(indices)
            char_locations = calculate_attention(
                image=img,
                attention_weights=attention_weights,
                index=max_length,
                vis=self.debug,
                log_dir=self.log_dir,
            )
            if max_length == 0:
                batch_char_locations = char_locations
            else:
                batch_char_locations = np.concatenate(
                    (batch_char_locations, char_locations), axis=1
                )
            max_length += 1

            del output

        translated_sentence = np.asarray(translated_sentence).T
        char_probs = np.asarray(char_probs).T
        char_probs_accumulate = np.multiply(char_probs, translated_sentence > 3)
        char_probs_accumulate = np.sum(char_probs, axis=-1) / (char_probs > 0).sum(-1)
        return (
            translated_sentence,
            char_probs,
            char_probs_accumulate,
            batch_char_locations,
        )

    def __call__(self, img, image_height=64, image_max_width=1024, image_min_width=64):
        if is_empty_image(img):
            return ""
        img, _ = process_image(
            img,
            image_height=image_height,
            image_max_width=image_max_width,
            image_min_width=image_min_width,
        )
        s, list_probs, probs, batch_locations = self.translate_onnx(img, self.session)
        sents = self.vocab.batch_decode(s.tolist())
        res = sents[0]
        probs = list_probs.squeeze(0).tolist()[1:-1]
        locations = batch_locations.squeeze(0).tolist()[:-1]
        img = img.squeeze(0).transpose(1, 2, 0) * 255
        res = remove_noise_text(img, res, probs, locations)
        return res
        pass



import cv2
import numpy as np


def find_cc_in_thresh_image(thresh_image):
    stats = cv2.connectedComponentsWithStats(thresh_image, 4, cv2.CV_32S)[2][1:]
    return stats[stats[:, 0].argsort()]


def bgr2thresh(bgr_image):
    gray_image = cv2.cvtColor(bgr_image, cv2.COLOR_BGR2GRAY)
    gray_image = cv2.GaussianBlur(gray_image, (3, 3), 0)
    _, thresh = cv2.threshold(
        gray_image, 150, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU
    )
    return thresh


def is_w_line(line_image, i):
    height, _ = line_image.shape
    stats = find_cc_in_thresh_image(line_image)
    for stat in stats:
        if stat[3] > 0.9 * height:  # h
            return True
    return False


def find_and_remove_w_line(thresh_cc_image, w_region=5):
    _, width = thresh_cc_image.shape
    start_id = None
    end_id = None
    list_start_id = []
    list_end_id = []
    width_of_line = 0
    max_width_of_line = 0
    for i in range(0, width):
        end_check = min(i + w_region, width)
        if is_w_line(thresh_cc_image[:, i:end_check], i):
            pos_y, _ = np.where(thresh_cc_image[:, i:end_check] > 0)
            min_y, max_y = np.min(pos_y), np.max(pos_y)
            for j in range(i, end_check):
                if thresh_cc_image[min_y, j] > 0 or thresh_cc_image[max_y, j] > 0:
                    start_id = j
                    list_start_id.append(j)
                    break
            for j in range(end_check - 1, i - 1, -1):
                if thresh_cc_image[min_y, j] > 0 or thresh_cc_image[max_y, j] > 0:
                    end_id = j
                    list_end_id.append(j)
                    break
        if start_id != None and end_id != None:
            width_of_line = end_id - start_id
            max_width_of_line = max(max_width_of_line, width_of_line)
    if len(list_start_id) > 0:
        if max_width_of_line == 1:
            max_width_of_line = min(max_width_of_line + 1, width)
        return list_start_id[0], max_width_of_line
    else:
        return None, 0


def rm_line_border(img_thresh):
    img = img_thresh.copy()
    height, width = img.shape
    if height >= 2.0 * width:
        return img

    erosion = cv2.dilate(img, kernel=np.ones((3, 10), np.uint8))
    if height > 100:
        check_h_size = 25
    else:
        check_h_size = 20

    if width / height >= 2.5:
        boxes = find_cc_in_thresh_image(erosion[:, 30 : width - 30])
        for box in boxes:
            if len(box) == 0:
                continue
            x_min, y_min, w, h, area = box
            if w > width * 0.4 and h <= 20 and (y_min <= 20 or y_min >= height - 20):
                img[y_min : y_min + h, :] = 0

    erosion = cv2.dilate(img, kernel=np.ones((3, 10), np.uint8))
    boxes = find_cc_in_thresh_image(erosion)
    big_box = False
    for box in boxes:
        x_min, y_min, w, h, area = box
        if (
            w > width * 0.4 and h <= 20 and (y_min <= 20 or y_min >= height - 20)
        ) and width / height >= 1.5:
            check_w_size = 25
        elif (
            w > width * 0.65 and h <= 15 and (y_min <= 15 or y_min >= height - 15)
        ) and width / height < 1.5:
            check_w_size = width
        else:
            if width / height >= 1.5:
                check_w_size = 25
            else:
                check_w_size = width
            big_box = True

        if big_box == False:
            erosion[y_min : y_min + h, :] = 0
            img[y_min : y_min + h, :] = 0

        x_line, w_line = find_and_remove_w_line(erosion[:, :check_w_size], w_region=10)
        if x_line != None:
            erosion[:, x_line : x_line + w_line] = 0
            img[:, x_line : x_line + w_line] = 0

        img_flip = cv2.flip(img, 1)
        erosion_flip = cv2.flip(erosion, 1)
        x_line, w_line = find_and_remove_w_line(
            erosion_flip[:, :check_w_size], w_region=10
        )
        if x_line != None:
            img_flip[:, x_line : x_line + w_line] = 0
            erosion_flip[:, x_line : x_line + w_line] = 0
            img = cv2.flip(img_flip, 1)
            erosion = cv2.flip(img_flip, 1)
        if big_box == True:
            boxes = find_cc_in_thresh_image(erosion)
            for box in boxes:
                x_min, y_min, w, h, area = box
                if (
                    (
                        (w > width * 0.4 and width / height >= 1.5)
                        or (w > width * 0.65 and width / height < 1.5)
                    )
                    and h <= check_h_size
                    and (y_min <= check_h_size or y_min >= height - check_h_size)
                ):
                    img[y_min : y_min + h, :] = 0
                if (
                    (y_min <= 3 or y_min >= height - 3)
                    and h > 0.9 * height
                    and (x_min <= 5 or x_min >= width - 5)
                    and w <= 10
                    and width < height
                ):
                    tmp = img.copy()
                    tmp[:, x_min : x_min + w] = 0
                    if np.sum(tmp[:, x_min : x_min + w]) < 10:
                        img = tmp.copy()
    return np.invert(img)


def check_pixel(white_bg_image, stats, img_shape):
    height, width = img_shape
    if np.sum(white_bg_image) > 0.99 * height * width * 255:
        if len(stats) == 1 and (stats[0][2] == 1 or stats[0][3] == 1):
            return True

    if np.sum(white_bg_image) > 0.98 * height * width * 255:
        if all(
            [
                (stat[1] > 0.97 * height or stat[1] < 0.02 * height)
                and stat[3] < height // 20
                for stat in stats
            ]
        ):
            return True

    if all(
        [
            stat[2] < 0.1 * height
            and stat[3] < 0.1 * height
            and 0.5 < stat[3] / stat[2] < 2.0
            for stat in stats
        ]
    ):
        return True

    if (
        all(
            [
                (
                    (stat[1] + stat[3] >= 0.8 * height or stat[0] == 0)
                    and stat[3] < height * 0.2
                )
                or (stat[1] <= 0.1 * height and stat[3] < 0.4 * height)
                for stat in stats
            ]
        )
        and stats[-1][0] - stats[0][0] >= 0.6 * width
        and len(stats) >= 5
    ):
        return True


def check_small_dot_cases(invert_white_bg_image, stats):
    if (invert_white_bg_image.shape[0] >= 21) and (
        invert_white_bg_image.shape[1] >= 21
    ):
        tmp_img = invert_white_bg_image.copy()[10:-10, 10:-10]
    else:
        tmp_img = invert_white_bg_image
    tmp_img = cv2.erode(tmp_img, np.ones((3, 3), np.uint8))
    vis = tmp_img.copy()
    for component in stats:
        x, y, w, h, _ = component
        if x < 10:
            w = w - x
        if y < 10:
            h = h - y
        if 1 <= w <= 15 and 1 <= h <= 15:
            x = max(x - 10, 0)
            y = max(y - 10, 0)
            tmp_img[y : y + h, x : x + w] = 0
            cv2.rectangle(
                vis, (x, y), (x + w, y + h), color=(255, 255, 255), thickness=1
            )
    if np.sum(tmp_img) <= 2000:
        return True


def check_empty_second(white_bg_image):
    height, width = white_bg_image.shape
    if np.mean(white_bg_image) == 255:
        return True

    invert_white_bg_image = np.invert(white_bg_image)
    stats = find_cc_in_thresh_image(invert_white_bg_image)

    if check_pixel(white_bg_image, stats, (height, width)):
        return True

    if check_small_dot_cases(invert_white_bg_image, stats):
        return True

    return False


def is_empty_image(bgr):
    gray = cv2.cvtColor(bgr, cv2.COLOR_BGR2GRAY)

    if np.sum(gray) > (0.999 * bgr.shape[0] * bgr.shape[1] * 255):
        return True

    thresh = bgr2thresh(bgr)
    if np.count_nonzero(thresh) > 0.4 * thresh.shape[0] * thresh.shape[1]:
        thresh = np.invert(thresh)

    if np.sum(gray * thresh) > np.sum(thresh) * 240 / 255:
        if np.average(gray) > 245:
            return True

    # check white second mode
    check_image = rm_line_border(thresh)
    if check_empty_second(check_image):  # it is white_bg !!!
        return True

    return False
